#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that defines that extends QColor functionality
"""

from __future__ import print_function, division, absolute_import

import math
import random

from Qt.QtCore import *
from Qt.QtGui import *

from tpDcc.libs.python import mathlib

_NUMERALS = '0123456789abcdefABCDEF'
_HEXDEC = {v: int(v, 16) for v in (x + y for x in _NUMERALS for y in _NUMERALS)}

# TODO: Change by enum
_LOWERCASE, _UPPERCASE = 'x', 'X'

REGEX_QCOLOR = r"^(?:(?:#[A-Fa-f0-9]{3})|(?:#[A-Fa-f0-9]{6})|(?:[a-zA-Z]+))$"
REGEX_FN_RGB = r"(^rgb\s*\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*\)$)"
REGEX_HEX_RGBA = r"^#[A-Fa-f0-9]{8}$"
REGEX_FN_RGBA = r"(^rgba?\s*\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*\)$)"


class Color(QColor, object):
    def __eq__(self, other):
        if other == self:
            return True
        elif isinstance(other, Color):
            return self.to_string() == other.to_string()
        else:
            return False

    @classmethod
    def from_color(cls, color):
        """
        Gets a string formatted color from a QColor
        :param color: QColor, color to parse
        :return: (str)
        """

        color = ('rgba(%d, %d, %d, %d)' % color.getRgb())
        return cls.from_string(color)

    @classmethod
    def from_string(cls, text_color):
        """
        Returns a (int, int, int, int) format color from a string format color
        :param text_color: str, string format color to parse
        :param alpha: int, alpha of the color
        :return: (int, int, int, int)
        """

        a = 255
        if string_is_hex(text_color):
            r, g, b = cls.rgb_from_hex(text_color)
        else:
            try:
                if text_color.startswith('rgba'):
                    r, g, b, a = text_color.replace('rgba(', '').replace(')', '').split(',')
                else:
                    r, g, b, a = text_color.replace('rgb(', '').replace(')', '').split(',')
            except ValueError:
                if text_color.startswith('rgba'):
                    r, g, b = text_color.replace('rgba(', '').replace(')', '').split(',')
                else:
                    r, g, b = text_color.replace('rgb(', '').replace(')', '').split(',')

        return cls(int(r), int(g), int(b), int(a))

    @classmethod
    def rgb_from_hex(cls, triplet):
        """
        Returns a RGB triplet from an hexadecimal value
        :param triplet: r,g,b Hexadecimal Color tuple
        """

        if triplet.startswith('#'):
            triplet = triplet[1:]

        if len(triplet) == 3:
            r, g, b = triplet[0] * 2, triplet[1] * 2, triplet[2] * 2
            return tuple([float(int(v, 16)) for v in (r, g, b)])

        return _HEXDEC[triplet[0:2]], _HEXDEC[triplet[2:4]], _HEXDEC[triplet[4:6]]

    @classmethod
    def hex_from_rgb(cls, rgb, lettercase=_LOWERCASE):
        """
        Returns a hexadecimal value from a triplet
        :param rgb: tuple(r,g,b) RGB Color tuple
        :param lettercase: LOWERCASE if you want to get a lowercase or UPPERCASE
        """

        return format(rgb[0] << 16 | rgb[1] << 8 | rgb[2], '06' + lettercase)

    @classmethod
    def rgb_to_hex(cls, rgb):
        """
        Returns RGB color from hexadecimal
        :param rgb:
        :return: str
        """

        return '#%02x%02x%02x' % rgb

    @classmethod
    def get_random_hex(cls, return_sign=True):
        """
        Returns a random HEX value
        :param return_sign: bool, True if you want to get color sign
        :return: str
        """

        def _random_int():
            return random.randint(0, 255)

        if return_sign:
            return '#%02X%02X%02X' % (_random_int(), _random_int(), _random_int())
        else:
            return '%02X%02X%02X' % (_random_int(), _random_int(), _random_int())

    @classmethod
    def get_random_rgb(cls):
        """
        Returns a random RGB color
        """

        hex = cls.get_random_hex(return_sign=False)
        return cls.rgb_from_hex(hex)

    @classmethod
    def hex_to_qcolor(cls, hex_color):

        """
        Converts a Hexadecimal color to QColor
        """

        r = int(hex_color[0:2], 16)
        g = int(hex_color[2:4], 16)
        b = int(hex_color[4:6], 16)
        return QColor(r, g, b)

    @classmethod
    def get_complementary_color(cls, color):

        """
        Returns the complementary color of the given color
        :param color: QColor
        """

        hsl_hue = color.hslHueF()
        hsl_saturation = color.hslSaturationF()
        lightness = color.lightnessF()
        lightness = 1.0 - lightness
        hsl_hue += 0.5
        if hsl_hue >= 1.0:
            hsl_hue -= 1.0
        return cls.fromHslF(hsl_hue, hsl_saturation, lightness)

    @classmethod
    def get_option_color(cls, option):
        """
        Returns QColor depending of the option passed as argument
        :param option: str, Option to get color
        """

        if option == 'Grey':
            color = cls.fromRgbF(0.4, 0.4, 0.4)
        elif option == 'Cancel':
            color = cls.fromRgbF(0.7, 0.5, 0.4)
        elif option == 'OK' or option == 'ok':
            color = cls.fromRgbF(0.5, 0.7, 0.8)
        elif option == 'Warning' or option == 'Warn':
            color = cls.fromRgbF(0.7, 0.2, 0.2)
        elif option == 'Collapse':
            color = cls.fromRgbF(0.15, 0.15, 0.15)
        elif option == 'Subtle':
            color = cls.fromRgbF(0.48, 0.48, 0.6)
        else:
            color = cls.fromRgbF(0.48, 0.48, 0.6)
        if option:
            color.ann = option + ' Color'
        return color

    @classmethod
    def expand_normalized_rgb(cls, normalized_rgb):
        return tuple([float(normalized_rgb[0]) * 255, float(normalized_rgb[1]) * 255, float(normalized_rgb[2]) * 255])

    @classmethod
    def normalized_rgb(cls, rgb):
        return tuple([float(rgb[0]) / 255, float(rgb[1]) / 255, float(rgb[2]) / 255])

    def to_string(self):
        """
        Returns the color with string format
        :return: str
        """

        return 'rgba(%d, %d, %d, %d)' % self.getRgb()

    def is_dark(self):
        """
        Return True if the color is considered dark (RGB < 125(mid grey)) or False otherwise
        :return: bool
        """

        return self.red() < 125 and self.green() < 125 and self.blue() < 125


# =================================================================================================================

DEFAULT_DARK_COLOR = Color(50, 50, 50, 255)
DEFAULT_LIGHT_COLOR = Color(180, 180, 180, 255)
BLACK = Color(0, 0, 0, 255)
GRAY = Color(110, 110, 110, 255)
RED = Color(255, 0, 0, 255)
GREEN = Color(0, 255, 0, 255)
BLUE = Color(0, 0, 255, 255)
YELLOW = Color(255, 255, 0, 255)
ORANGE = Color(209, 84, 0, 255)
MAGENTA = Color(1.0, 0.0, 1.0)
CYAN = Color(0.0, 1.0, 1.0)
WHITE = Color(1.0, 1.0, 1.0)
DARK_GRAY = Color(60, 60, 60, 255)
DARK_RED = Color(0.75, 0.0, 0.0)
DARK_GREEN = Color(0.0, 0.75, 0.0)
DARK_BLUE = Color(0.0, 0.0, 0.75)
DARK_YELLOW = Color(0.75, 0.75, 0.0)
DARK_MAGENTA = Color(0.75, 0.0, 0.75)
DARK_CYAN = Color(0.0, 0.75, 0.75)
LIGHT_GRAY = Color(0.75, 0.75, 0.75)
LIGHT_RED = Color(1.0, 0.25, 0.25)
LIGHT_GREEN = Color(0.25, 1.0, 0.25)
LIGHT_BLUE = Color(0.25, 0.25, 1.0)
LIGHT_YELLOW = Color(1.0, 1.0, 0.25)
LIGHT_MAGENTA = Color(1.0, 0.25, 1.0)
LIGHT_CYAN = Color(0.25, 1.0, 1.0)

# =================================================================================================================


def string_is_hex(color_str):
    """
    Returns whether or not given string is a valid hexadecimal color
    :param color_str: str
    :return: bool
    """

    if color_str.startswith('#'):
        color_str = color_str[1:]
    hex_regex1 = QRegExp('^[0-9A-F]{3}$', Qt.CaseInsensitive)
    hex_regex2 = QRegExp('^[0-9A-F]{6}$', Qt.CaseInsensitive)
    hex_regex3 = QRegExp('^[0-9A-F]{8}$', Qt.CaseInsensitive)
    if hex_regex1.exactMatch(color_str) or hex_regex2.exactMatch(color_str) or hex_regex3.exactMatch(color_str):
        return True

    return False


def convert_2_hex(color):
    """
    Converts given color to hexadecimal value
    :param color:
    :return: str
    """

    if isinstance(color, (str, unicode)) and string_is_hex(color):
        return color

    hex = '#'
    for var in color:
        var = format(var, 'x')
        if len(var) == 1:
            hex += '0' + str(var)
        else:
            hex += str(var)

    if len(hex) == 9:
        hex = '#{}{}'.format(hex[-2:], hex[1:-2])

    return hex


def generate_color(primary_color, index):
    """
    Generates a new color from the given one and with given index (between 1 and 10)
    https://github.com/phenom-films/dayu_widgets/blob/master/dayu_widgets/utils.py
    :param primary_color: base color (RRGGBB)
    :param index: color step from 1 (light) to 10 (dark)
    :return: out color Color
    """

    hue_step = 2
    saturation_step = 16
    saturation_step2 = 5
    brightness_step1 = 5
    brightness_step2 = 15
    light_color_count = 5
    dark_color_count = 4

    def _get_hue(color, i, is_light):
        h_comp = color.hue()
        if 60 <= h_comp <= 240:
            hue = h_comp - hue_step * i if is_light else h_comp + hue_step * i
        else:
            hue = h_comp + hue_step * i if is_light else h_comp - hue_step * i
        if hue < 0:
            hue += 359
        elif hue >= 359:
            hue -= 359
        return hue / 359.0

    def _get_saturation(color, i, is_light):
        s_comp = color.saturationF() * 100
        if is_light:
            saturation = s_comp - saturation_step * i
        elif i == dark_color_count:
            saturation = s_comp + saturation_step
        else:
            saturation = s_comp + saturation_step2 * i
        saturation = min(100.0, saturation)
        if is_light and i == light_color_count and saturation > 10:
            saturation = 10
        saturation = max(6.0, saturation)
        return round(saturation * 10) / 1000.0

    def _get_value(color, i, is_light):
        v_comp = color.valueF()
        if is_light:
            return min((v_comp * 100 + brightness_step1 * i) / 100, 1.0)
        return max((v_comp * 100 - brightness_step2 * i) / 100, 0.0)

    light = index <= 6
    hsv_color = Color(primary_color) if isinstance(primary_color, basestring) else primary_color
    index = light_color_count + 1 - index if light else index - light_color_count - 1
    return Color(QColor.fromHsvF(
        _get_hue(hsv_color, index, light),
        _get_saturation(hsv_color, index, light),
        _get_value(hsv_color, index, light)
    )).name()


def string_from_color(color, alpha):
    if not alpha or color.alpha() == 255:
        return color.name()
    return color.name() + '{}'.format(color.alpha())
    # color.name()+QStringLiteral("%1").arg(color.alpha(), 2, 16, QChar('0'));


def color_from_string(string, alpha):
    xs = string.strip()
    regex = QRegExp(REGEX_QCOLOR)
    match = regex.exactMatch(xs)
    if match:
        return QColor(xs)

    regex = QRegExp(REGEX_FN_RGB)
    match = regex.exactMatch(xs)
    if match:
        captured_texts = regex.capturedTexts()
        return QColor(int(captured_texts[-3]), int(captured_texts[-2]), int(captured_texts[-1]))

    if alpha:
        regex = QRegExp(REGEX_HEX_RGBA)
        match = regex.exactMatch(xs)
        if match:
            return QColor(_HEXDEC[xs[1:3]], _HEXDEC[xs[3:5]], _HEXDEC[xs[5:7]], _HEXDEC[xs[7:9]])

        regex = QRegExp(REGEX_FN_RGBA)
        match = regex.exactMatch(xs)
        if match:
            captured_texts = regex.capturedTexts()
            return QColor(
                int(captured_texts[-4]), int(captured_texts[-3]), int(captured_texts[-2]), int(captured_texts[-1]))

    return QColor()


def color_chroma_float(color):
    max_value = max(color.redF(), max(color.greenF(), color.blueF()))
    min_value = min(color.redF(), min(color.greenF(), color.blueF()))

    return max_value - min_value


def color_luma_float(color):
    return 0.30 * color.redF() + 0.59 * color.greenF() + 0.11 * color.blueF()


def color_from_lch(hue, chroma, luma, alpha=1):
    h1 = hue * 6
    x = chroma * (1 - abs(math.fmod(h1, 2) - 1))
    col = QColor()
    if 0 <= h1 < 1:
        col = QColor.fromRgbF(chroma, x, 0)
    elif h1 < 2:
        col = QColor.fromRgbF(x, chroma, 0)
    elif h1 < 3:
        col = QColor.fromRgbF(0, chroma, x)
    elif h1 < 4:
        col = QColor.fromRgbF(0, x, chroma)
    elif h1 < 5:
        col = QColor.fromRgbF(x, 0, chroma)
    elif h1 < 6:
        col = QColor.fromRgbF(chroma, 0, x)

    m = luma - color_luma_float(col)

    return QColor.fromRgbF(
        mathlib.clamp(col.redF() + m, 0.0, 1.0),
        mathlib.clamp(col.greenF() + m, 0.0, 1.0),
        mathlib.clamp(col.blueF() + m, 0.0, 1.0),
        alpha
    )


def rainbow_lch(hue):
    return color_from_lch(hue, 1, 1)


def rainbow_hsv(hue):
    return QColor.fromHsvF(hue, 1, 1)


def color_lightnes_float(color):
    return (max(color.redF(), max(
        color.greenF(), color.blueF())) + min(color.redF(), min(color.greenF(), color.blueF()))) / 2


def color_hsl_saturation_float(color):
    chroma_value = color_chroma_float(color)
    lightness_value = color_lightnes_float(color)
    if qFuzzyCompare(lightness_value + 1, 1) or qFuzzyCompare(lightness_value + 1, 2):
        return 0

    return chroma_value / (1 - abs(2 * lightness_value - 1))


def color_from_hsl(hue, sat, lig, alpha):
    chroma = (1 - abs(2 * lig - 1)) * sat
    h1 = hue * 6
    x = chroma * (1 - abs(math.fmod(h1, 2) - 1))
    col = QColor()
    if 0 <= h1 < 1:
        col = QColor.fromRgbF(chroma, x, 0)
    elif h1 < 2:
        col = QColor.fromRgbF(x, chroma, 0)
    elif h1 < 3:
        col = QColor.fromRgbF(0, chroma, x)
    elif h1 < 4:
        col = QColor.fromRgbF(0, x, chroma)
    elif h1 < 5:
        col = QColor.fromRgbF(x, 0, chroma)
    elif h1 < 6:
        col = QColor.fromRgbF(chroma, 0, x)

    m = lig - chroma / 2

    return QColor.fromRgbF(
        mathlib.clamp(col.redF() + m, 0.0, 1.0),
        mathlib.clamp(col.greenF() + m, 0.0, 1.0),
        mathlib.clamp(col.blueF() + m, 0.0, 1.0),
        alpha
    )



# This file is part of the PySide project.
#
# Copyright (C) 2009-2011 Nokia Corporation and/or its subsidiary(-ies).
# Copyright (C) 2010 Riverbank Computing Limited.
# Copyright (C) 2009 Torsten Marek
#
# Contact: PySide team <pyside@openbossa.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301 USA

from pysideuic.exceptions import NoSuchWidgetError


def invoke(driver):
    """ Invoke the given command line driver.  Return the exit status to be
    passed back to the parent process.
    """

    exit_status = 1

    try:
        exit_status = driver.invoke()

    except IOError, e:
        driver.on_IOError(e)

    except SyntaxError, e:
        driver.on_SyntaxError(e)

    except NoSuchWidgetError, e:
        driver.on_NoSuchWidgetError(e)

    except Exception, e:
        driver.on_Exception(e)

    return exit_status



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom tray balloon
"""

from __future__ import print_function, division, absolute_import

from Qt.QtWidgets import *


class TrayMessage(QWidget, object):

    def __init__(self, parent=None):
        super(TrayMessage, self).__init__(parent=parent)

        self._tools_icon = None

        self.tray_icon_menu = QMenu(self)
        self.tray_icon = QSystemTrayIcon(self)
        # self.tray_icon.setIcon(self._tools_icon)
        self.tray_icon.setToolTip('Tray')
        self.tray_icon.setContextMenu(self.tray_icon_menu)

        if not QSystemTrayIcon.isSystemTrayAvailable():
            raise OSError('Tray Icon is not available!')

        self.tray_icon.show()

    def show_message(self, title, msg):
        try:
            self.tray_icon.showMessage(title, msg, self._tools_icon)
        except Exception:
            self.tray_icon.showMessage(title, msg)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains main library object implementation
"""

from __future__ import print_function, division, absolute_import

import os
import time
import copy
from collections import OrderedDict

from Qt.QtCore import *

from tpDcc.libs import qt
from tpDcc.libs.python import decorators, path as path_utils
from tpDcc.libs.qt.widgets.library import consts, utils


class Library(QObject, object):

    Name = consts.LIBRARY_DEFAULT_NAME

    Fields = [
        'icon',
        'name',
        'path',
        'type',
        'folder',
        'category'
    ]

    SortFields = [
        'name',
        'path',
        'type',
        'folder',
        'category'
    ]

    GroupFields = [
        'type',
        'category'
    ]

    dataChanged = Signal()
    searchStarted = Signal()
    searchFinished = Signal()
    searchTimeFinished = Signal()

    def __init__(self, path=None, library_window=None, *args):

        self._path = path
        self._mtime = None
        self._data = dict()
        self._items = list()
        self._fields = list()
        self._sort_by = list()
        self._group_by = list()
        self._results = list()
        self._grouped_results = dict()
        self._queries = dict()
        self._global_queries = dict()
        self._search_time = 0
        self._search_enabled = True
        self._library_window = library_window

        super(Library, self).__init__(*args)

        self.set_path(path)
        self.set_dirty(True)

    @decorators.abstractmethod
    def data_path(self):
        """
        Returns path where library data base is located
        :return: str
        """

        raise NotImplementedError('Library data_path() not implemented!')

    @staticmethod
    def match(data, queries):
        """
        Match the given data with the given queries
        :param data: dict
        :param queries: list(dict)
        :return: list
        """

        matches = list()

        for query in queries:
            filters = query.get('filters')
            operator = query.get('operator', 'and')
            if not filters:
                continue

            match = False
            for key, cond, value in filters:
                if key == '*':
                    item_value = str(data)
                else:
                    item_value = data.get(key)

                if isinstance(value, (unicode, str)):
                    value = value.lower()
                if isinstance(item_value, (unicode, str)):
                    item_value = item_value.lower()
                if not item_value:
                    match = False
                elif cond == 'contains':
                    match = value in item_value
                elif cond == 'not_contains':
                    match = value not in item_value
                elif cond == 'is':
                    match = value == item_value
                elif cond == 'not':
                    match = value != item_value
                elif cond == 'startswith':
                    match = item_value.startswith(value)

                if operator == 'or' and match:
                    break
                if operator == 'and' and not match:
                    break

            matches.append(match)

        return all(matches)

    @staticmethod
    def sorted(items, sort_by):
        """
        Return the given data sorted using the sorty_by argument
        :param items: list(LibraryItem)
        :param sort_by: list(str)
        :return: list(LibraryItem)
        """

        qt.logger.debug('Sort by: {}'.format(sort_by))
        t = time.time()
        for field in reversed(sort_by):
            tokens = field.split(':')
            reverse = False
            if len(tokens) > 1:
                field = tokens[0]
                reverse = tokens[1] != 'asc'

            def sort_key(item):
                default = False if reverse else ''
                return item.item_data().get(field, default)

            items = sorted(items, key=sort_key, reverse=reverse)
        qt.logger.debug('Sort items took {}'.format(time.time() - t))

        return items

    @staticmethod
    def group_items(items, fields):
        """
        Group the given items by the given field
        :param items: list(LibraryItem)
        :param fields: list(str)
        :return: dict
        """

        qt.logger.debug('Group by: {}'.format(fields))

        # TODO: Implement support for multiple grups not only top level group

        if fields:
            field = fields[0]
        else:
            return {'None': items}

        t = time.time()
        results_ = dict()
        tokens = field.split(':')

        reverse = False
        if len(tokens) > 1:
            field = tokens[0]
            reverse = tokens[1] != 'asc'

        for item in items:
            value = item.item_data().get(field)
            if value:
                results_.setdefault(value, list())
                results_[value].append(item)

        groups = sorted(results_.keys(), reverse=reverse)

        results = OrderedDict()
        for group in groups:
            results[group] = results_[group]

        qt.logger.debug('Group Items Took {}'.format(time.time() - t))

        return results

    def manager(self):
        """
        Returns data manager used by this library
        :return: LibraryDataManager
        """

        if not self._library_window:
            return None

        return self._library_window.manager()

    def name(self):
        """
        Returns the name of the library
        :return: str
        """

        return self.Name

    def path(self):
        """
        Returns the path where library is located
        :return: str
        """

        return self._path

    def set_path(self, path):
        """
        Sets path where muscle data is located
        :param path: str
        """

        self._path = path

    def settings(self):
        """
        Returns the stetins for the data
        :return: dict
        """

        return {
            'sortBy': self.sort_by(),
            'groupBy': self.group_by()
        }

    def set_settings(self, settings):
        """
        Set the settings for the data
        :param settings: dict
        """

        value = settings.get('sortBy')
        if value:
            self.set_sort_by(value)
        value = settings.get('groupBy')
        if value:
            self.set_group_by(value)

    def distinct(self, field, queries=None, sort_by='name'):
        """
        Returns all values for the given field
        :param field: str
        :param queries: variant, None or list(dict)
        :param sort_by: str
        :return: list
        """

        results = dict()
        queries = queries or list()
        queries.extend(self._global_queries.values())

        items = self.create_items()
        if not items:
            return results

        for item in items:
            value = item.item_data().get(field)
            if value:
                results.setdefault(value, {'count': 0, 'name': value})
                match = self.match(item.item_data(), queries)
                if match:
                    results[value]['count'] += 1

        def sort_key(facet):
            return facet.get(sort_by)

        return sorted(results.values(), key=sort_key)

    def mtime(self):
        """
        Returns when the data was last modified
        :return: float or None
        """

        path = self.data_path()
        mtime = None
        if os.path.exists(path):
            mtime = os.path.getmtime(path)

        return mtime

    def sort_by(self):
        """
        Return the list of fields to sorty by
        :return: list(str)
        """

        return self._sort_by

    def set_sort_by(self, fields):
        """
        Set the list of fields to group by
        >>> set_sorty_by(['name:asc', 'type:asc'])
        :param fields: list(str)
        """

        self._sort_by = fields

    def group_by(self):
        """
        Return the list of fields to group by
        :return: list(str)
        """

        return self._group_by

    def set_group_by(self, fields):
        """
        Set the list of fields to group by
        >>> set_group_by(['name:asc', 'type:asc'])
        :param fields: list(str)
        """

        self._group_by = fields

    def fields(self):
        """
        Returns all the fields for the library
        :return: list(str)
        """

        return self._fields

    def is_dirty(self):
        """
        Returns whether the data has changed on disk or not
        :return: bool
        """

        return not self._items or self._mtime != self.mtime()

    def set_dirty(self, value):
        """
        Updates the model object with the current data timestamp
        :param value: bool
        """

        if value:
            self._mtime = None
        else:
            self._mtime = self.mtime()

    def read(self):
        """
        Read the data from disk and returns it a dictionary object
        :return: dict
        """

        if not self.path():
            qt.logger.info('No path set for reading the data from disk')
            return self._data

        if self.is_dirty():
            self._data = utils.read_json(self.data_path())
            self.set_dirty(False)

        return self._data

    def save(self, data):
        """
        Write the given data dict object to the data on disk
        :param data: dict
        """

        if not self.path():
            qt.logger.info('No path set for saving the data to disk')

        utils.save_json(self.data_path(), data)
        self.set_dirty(True)

    def is_search_enabled(self):
        """
        Returns whether search functionality is enabled or not
        :return: bool
        """

        return self._search_enabled

    def set_search_enabled(self, flag):
        """
        Sets whether search functionality is enabled or not
        :param flag: bool
        """

        self._search_enabled = flag

    def recursive_depth(self):
        """
        Return the recursive search depth
        :return: int
        """

        return consts.DEFAULT_RECURSIVE_DEPTH

    def add_item(self, item):
        """
        Add the given item to the library data
        :param item: LibraryItem
        """

        self.save_item_data([item])

    def add_items(self, items):
        """
        Add the given items to the library data
        :param items: list(LibraryItem)
        """

        self.save_item_data(items)

    def update_item(self, item):
        """
        Update the given item in the library data
        :param item: LibraryItem
        """

        self.save_item_data([item])

    def save_item_data(self, items, emit_data_changed=True):
        """
        Add the given items to the library data
        :param items: list(LibraryItem)
        :param emit_data_changed: bool
        """

        qt.logger.debug('Saving Items: {}'.format(items))

        data_ = self.read()
        for item in items:
            path = item.path()
            data = item.item_data()
            data_.setdefault(path, {})
            data_[path].update(data)

        self.save(data_)

        if emit_data_changed:
            self.search()
            self.dataChanged.emit()

    def load_item_data(self, items):
        """
        load the item data from the library data to the given items
        :param items: list(LibraryItem)
        """

        qt.logger.debug('Loading item data: {}'.format(items))

        data = self.read()
        for item in items:
            key = item.id()
            if key in data:
                item.set_item_data(data[key])

    def update_paths(self, paths, data):
        """
        Updates the given paths with the given data in the data
        :param paths: list(str)
        :param data: dict
        """

        data_ = self.read()
        paths = path_utils.normalize_paths(paths)
        for path in paths:
            if path in data_:
                data_[path].update(data)
            else:
                data_[path] = data

        self.save(data_)

    def add_paths(self, paths, data=None):
        """
        Add the given paths and the given data to the data
        :param paths: list(str)
        :param data: dict or None
        """

        data = data or dict()
        self.update_paths(paths, data)

    def copy_path(self, source, target):
        """
        Copy the given source path to the given target path
        :param source: str
        :param target:str
        :return: str
        """

        self.add_paths([target])
        return target

    def rename_item(self, item, target, extension=None):
        """
        Renames given item
        :param item: LibraryItem
        :param target: str
        :param extension: str
        """

        extension = extension or item.extension()
        if target and extension not in target:
            target += extension

        source = item.path()

        target = utils.rename_path(source, target)
        self.rename_path(source, target)

        item.set_path(target)
        item.save_item_data()

        return target

    def rename_path(self, source, target):
        """
        Renames the suorce path to the given target name
        :param source: str
        :param target: str
        :return: str
        """

        utils.rename_path_in_file(self.data_path(), source, target)
        self.set_dirty(True)

        return target

    def remove_path(self, path):
        """
        Removes the given path from the data
        :param path: str
        """

        self.remove_paths([path])

    def remove_paths(self, paths):
        """
        Removes the given paths from the data
        :param paths: list(str)
        """

        data = self.read()
        paths = path_utils.normalize_paths(paths)
        for path in paths:
            if path in data:
                del data[path]

        self.save(data)

    def find_items(self, queries):
        """
        Get the items that match the given queries
        :param queries: list(dict)
        :return: list(LibraryItem)
        """

        fields = list()
        results = list()

        queries = copy.copy(queries)
        queries.extend(self._global_queries.values())

        items = self.create_items()
        if not items:
            return results

        for item in items:
            match = self.match(item.item_data(), queries)
            if match:
                results.append(item)
            fields.extend(item.item_data().keys())

        self._fields = list(set(fields))

        if self.sort_by():
            results = self.sorted(results, self.sort_by())

        return results

    def find_items_by_path(self, item_path):
        """
        Returns item with given path
        :param item_path: str
        """

        found_items = list()

        items = self.create_items()
        if not items:
            return found_items

        for item in items:
            if item.path() == item_path:
                found_items.append(item)

        return found_items

    def queries(self, exclude=None):
        """
        Return all queries for the data excluding the given ones
        :param exclude: list(str) or None
        :return: list(dict)
        """

        queries = list()
        exclude = exclude or list()

        for query in self._queries.values():
            if query.get('name') not in exclude:
                queries.append(query)

        return queries

    def query_exists(self, name):
        """
        Check if the given queryh name exists
        :param name: str
        :return: bool
        """

        return name in self._queries

    def add_to_global_query(self, query):
        """
        Add a global query to library
        :param query: dict
        """

        self._global_queries[query['name']] = query

    def add_query(self, query):
        """
        Add a search query to the library
        >>> add_query({
        >>>    'name': 'Test Query',
        >>>    'operator': 'or',
        >>>    'filters': [
        >>>        ('folder', 'is', '/lib/proj/test'),
        >>>        ('folder', 'startswith', '/lib/proj/test'),
        >>>    ]
        >>>})
        :param query: dict
        """

        self._queries[query['name']] = query

    def remove_query(self, name):
        """
        Remove the query with the given name
        :param name: str
        """

        if name in self._queries:
            del self._queries[name]

    def search(self):
        """
        Run a search using the queries added to library data
        """

        if not self.is_search_enabled():
            return

        t = time.time()
        qt.logger.debug('Searching items ...')
        self.searchStarted.emit()
        self._results = self.find_items(self.queries())
        self._grouped_results = self.group_items(self._results, self.group_by())
        self.searchFinished.emit()
        self._search_time = time.time() - t
        self.searchTimeFinished.emit()
        qt.logger.debug('Search time: {}'.format(self._search_time))

    def results(self):
        """
        Return the items found after a search is executed
        :return: list(LibraryItem)
        """

        return self._results

    def grouped_results(self):
        """
        Return the results grouped after a search is executed
        :return: dict
        """

        return self._grouped_results

    def search_time(self):
        """
        Return the time taken to run a search
        :return: float
        """

        return self._search_time

    def sync(self, percent_callback=lambda message, percent: None):
        """
        Sync the file sytem wit hthe library data
        """

        if not self.path():
            qt.logger.warning('No path set for syncing data')
            return

        data = self.read()

        for path in data.keys():
            if not os.path.exists(path):
                del data[path]

        depth = self.recursive_depth()
        items = list(self._library_window.manager().find_items(self.path(), depth=depth))
        count = len(items)
        for i, item in enumerate(items):
            percent = (float(i + 1) / float(count))
            percent_callback('', percent)
            path = item.path()
            item_data = data.get(path, {})
            item_data.update(item.item_data())
            data[path] = item_data

        percent_callback('Post Sync', -1)
        self.post_sync(data)

        percent_callback('Saving Cache', -1)
        self.save(data)

        self.dataChanged.emit()

    def post_sync(self, data):
        """
        This function is called after a data sync, but before save and dataChanged signal is emitted
        :param data: dict
        """

        pass

    def create_items(self):
        """
        Create all teh items for the library model
        :return: list(LibraryItem)
        """

        if self.is_dirty():
            data = self.read()
            # if not data:
            #     return
            paths = data.keys()
            items = self._library_window.manager().items_from_paths(
                paths=paths,
                library=self,
                library_window=self._library_window
            )
            self._items = list(items)
            self.load_item_data(self._items)

        return self._items

    def clear(self):
        """
        Clear all the item data
        """

        self._items = list()
        self._results = list()
        self._grouped_results = dict()
        self.dataChanged.emit()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for overlay widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

from tpDcc.libs.qt.core import dialog


class OverlayWidget(dialog.Dialog, object):
    """
    Invisible overlay widget used to capture mouse events among other things
    """

    widgetMousePress = Signal(object)
    widgetMouseMove = Signal(object)
    widgetMouseRelease = Signal(object)

    OVERLAY_ACTIVE_KEY = Qt.AltModifier
    _PRESSED = False

    def __init__(self, parent):
        super(OverlayWidget, self).__init__(parent=parent, show_on_initialize=False)

        self.set_debug_mode(False)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def ui(self):
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
        self.setMouseTracking(True)
        self.update()

    def mousePressEvent(self, event):
        """
        Overrides base QDialog mousePressEvent
        :param event:
        :return:
        """

        if not QApplication.keyboardModifiers() == self.OVERLAY_ACTIVE_KEY:
            event.ignore()
            return

        self.widgetMousePress.emit(event)
        self._PRESSED = True
        self.update()

    def enterEvent(self, event):
        """
        Overrides base QDialog enterEvent
        :param event:
        :return:
        """

        if not self._PRESSED:
            self.hide()

        event.ignore()

    def mouseMoveEvent(self, event):
        """
        Overrides base QDialog mouseMoveEvent
        :param event:
        :return:
        """

        pass

    def mouseReleaseEvent(self, event):
        """
        Overrides base QDialog mouseReleaseEvent
        :param event:
        :return:
        """

        self.widgetMouseRelease.emit(event)
        self._PRESSED = False
        self.update()

    def keyReleaseEvent(self, event):
        """
        Overrides base QDialog mousePressEvent
        :param event:
        :return:
        """

        self.hide()

    def update(self):
        """
        Overrides base QDialog update function
        Updates geometry to match parents geometry
        :return:
        """

        self.setGeometry(0, 0, self.parent().geometry().width(), self.parent().geometry().height())

    def show(self, *args, **kwargs):
        """
        Overrides base QDialog show
        :param event:
        :return:
        """

        self._PRESSED = True
        super(OverlayWidget, self).show(*args, **kwargs)
        self.update()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def set_debug_mode(self, flag):
        """
        Enables or disable debug mode by making window semi red transparent. Useful to know where the widget is located
        :param flag: bool
        """

        if flag:
            self.setStyleSheet('background-color: #88ff0000;')
        else:
            self.setStyleSheet('background-color: transparent;')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains main window widget used in libraries
"""

from __future__ import print_function, division, absolute_import

import os
import re
import time
import operator
from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc.libs.python import decorators, path as path_utils
from tpDcc.libs import qt
from tpDcc.libs.qt.core import base, icon, menu, qtutils, animation, window
from tpDcc.libs.qt.widgets import stack, messagebox, action
from tpDcc.libs.qt.widgets.library import consts, utils, library, viewer, widgets

if tp.is_maya():
    from tpDcc.dccs.maya.core import decorators as maya_decorators
    show_wait_cursor_decorator = maya_decorators.show_wait_cursor
    show_arrow_cursor_decorator = maya_decorators.show_arrow_cursor
else:
    show_wait_cursor_decorator = decorators.empty_decorator
    show_arrow_cursor_decorator = decorators.empty_decorator


class GlobalSignal(QObject):
    """
    Signals that are triggered by all library instances
    """

    debugModeChanged = Signal(object, object)
    folderSelectionChanged = Signal(object, object)


class PreviewFrame(QFrame):
    pass


class SidebarFrame(QFrame):
    pass


class LibraryWindow(window.BaseWindow, object):
    LIBRARY_CLASS = library.Library
    VIEWER_CLASS = viewer.LibraryViewer
    SEARCH_WIDGET_CLASS = widgets.LibrarySearchWidget
    STATUS_WIDGET_CLASS = widgets.LibraryStatusWidget
    MENUBAR_WIDGET_CLASS = widgets.LibraryToolbarWidget
    SIDEBAR_WIDGET_CLASS = widgets.LibrarySidebarWidget
    SORTBY_MENU_CLASS = widgets.SortByMenu
    GROUPBY_MENU_CLASS = widgets.GroupByMenu
    FILTERBY_MENU_CLASS = widgets.FilterByMenu
    PROGRESS_BAR_VISIBLE = consts.PROGRESS_BAR_VISIBLE

    globalSignal = GlobalSignal()
    loaded = Signal()
    lockChanged = Signal(object)
    debugModeChanged = Signal(object)
    itemRenamed = Signal(str, str)
    itemSelectionChanged = Signal(object)
    folderRenamed = Signal(str, str)
    folderSelectionChanged = Signal(object)

    def __init__(self, parent=None, name='', path='', library_icon_path=None, allow_non_path=False, **kwargs):

        self._items = list()
        self._name = name or consts.LIBRARY_DEFAULT_NAME
        self._is_debug = False
        self._is_locked = False
        self._preview_widget = None
        self._new_item_widget = None
        self._progress_bar = None
        self._current_item = None
        self._library = None
        self._refresh_enabled = False
        self._library_icon_path = library_icon_path
        self._allow_non_path = allow_non_path
        self._superusers = None
        self._lock_reg_exp = None
        self._unlock_reg_exp = None

        self._trash_enabled = consts.TRASH_ENABLED
        self._recursive_search_enabled = consts.DEFAULT_RECURSIVE_SEARCH_ENABLED

        self._items_hidden_count = 0
        self._items_visible_count = 0

        self._is_trash_folder_visible = False
        self._sidebar_widget_visible = True
        self._preview_widget_visible = True
        self._status_widget_visible = True
        self._new_item_widget_visible = False

        super(LibraryWindow, self).__init__(parent=parent)
        # super(LibraryWindow, self).__init__(name=name, parent=parent, show_dragger=False, auto_load=False, **kwargs)

        self.set_refresh_enabled(True)

        self.update_view_button()
        self.update_filters_button()
        self.update_preview_widget()

        if path:
            self.set_path(path)

    # ============================================================================================================
    # ABSTRACT
    # ============================================================================================================

    @decorators.abstractmethod
    def manager(self):
        """
        Returns library managed used by this window
        Must be implemented in child classes
        :return: LibraryManager
        """

        raise NotImplementedError('LibraryWindow manager() function is not implemented!')

    # ============================================================================================================
    # OVERLOADING FUNCTIONS
    # ============================================================================================================

    def ui(self):
        super(LibraryWindow, self).ui()

        self.setMinimumWidth(5)
        self.setMinimumHeight(5)

        self.stack = stack.SlidingStackedWidget()
        self.main_layout.addWidget(self.stack)

        lib = self.LIBRARY_CLASS(library_window=self)
        lib.dataChanged.connect(self.refresh)
        lib.searchTimeFinished.connect(self._on_search_finished)

        self._sidebar_frame = SidebarFrame(self)
        sidebar_frame_lyt = QVBoxLayout(self)
        sidebar_frame_lyt.setContentsMargins(0, 1, 0, 0)
        self._sidebar_frame.setLayout(sidebar_frame_lyt)

        self._new_item_frame = PreviewFrame(self)
        self._new_item_frame.setMinimumWidth(5)
        new_item_lyt = QVBoxLayout()
        new_item_lyt.setSpacing(0)
        new_item_lyt.setContentsMargins(0, 0, 0, 0)
        self._new_item_frame.setLayout(new_item_lyt)

        self._preview_frame = PreviewFrame(self)
        self._preview_frame.setMinimumWidth(5)
        preview_frame_lyt = QVBoxLayout()
        preview_frame_lyt.setSpacing(0)
        preview_frame_lyt.setContentsMargins(0, 0, 0, 0)
        self._preview_frame.setLayout(preview_frame_lyt)

        self._viewer = self.VIEWER_CLASS(self)
        self._viewer.setContextMenuPolicy(Qt.CustomContextMenu)

        self._sort_by_menu = self.SORTBY_MENU_CLASS(self)
        self._group_by_menu = self.GROUPBY_MENU_CLASS(self)
        self._filter_by_menu = self.FILTERBY_MENU_CLASS(self)
        self._status_widget = self.STATUS_WIDGET_CLASS(self)
        self._menubar_widget = self.MENUBAR_WIDGET_CLASS(self)

        self._sidebar_widget = self.SIDEBAR_WIDGET_CLASS(self)
        self._sidebar_widget.setContextMenuPolicy(Qt.CustomContextMenu)

        sidebar_frame_lyt.addWidget(self._sidebar_widget)

        self._search_widget = self.SEARCH_WIDGET_CLASS(self)

        self._splitter = QSplitter(Qt.Horizontal, self)
        self._splitter.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Expanding)
        self._splitter.setHandleWidth(2)
        self._splitter.setChildrenCollapsible(False)

        self._splitter.addWidget(self._sidebar_frame)
        self._splitter.addWidget(self._viewer)
        self._splitter.addWidget(self._preview_frame)

        self._splitter.setStretchFactor(0, False)
        self._splitter.setStretchFactor(1, True)
        self._splitter.setStretchFactor(2, False)

        base_widget = QWidget()
        base_layout = QVBoxLayout()
        base_layout.setContentsMargins(0, 0, 0, 0)
        base_layout.setSpacing(0)
        base_widget.setLayout(base_layout)
        base_layout.addWidget(self._menubar_widget)
        base_layout.addWidget(self._splitter)
        base_layout.addWidget(self._status_widget)

        self.stack.addWidget(base_widget)
        self.stack.addWidget(self._new_item_frame)

        self._sort_by_menu.set_library(lib)
        self._group_by_menu.set_library(lib)
        self._filter_by_menu.set_library(lib)
        self._viewer.set_library(lib)
        self._search_widget.set_library(lib)
        self._sidebar_widget.set_library(lib)
        self.set_library(lib)

        self._setup_menubar()

    def setup_signals(self):
        self._viewer.setContextMenuPolicy(Qt.CustomContextMenu)
        self._viewer.itemMoved.connect(self._on_item_moved)
        self._viewer.itemDropped.connect(self._on_item_dropped)
        self._viewer.itemSelectionChanged.connect(self._on_item_selection_changed)
        self._viewer.customContextMenuRequested.connect(self._on_show_items_context_menu)

        self._sidebar_widget.itemDropped.connect(self._on_item_dropped)
        self._sidebar_widget.itemSelectionChanged.connect(self._on_folder_selection_changed)
        self._sidebar_widget.customContextMenuRequested.connect(self._on_show_folder_menu)

        self.folderSelectionChanged.connect(self.update_lock)

    def _setup_menubar(self):
        icon_color = self.icon_color()
        name = 'New Item'
        icon = tp.ResourcesMgr().icon('plus', theme='black')
        icon.set_color(icon_color)
        tip = 'Add a new item to the selected folder'
        self.add_menubar_action(name, icon, tip, callback=self._on_show_new_menu)

        self._menubar_widget.addWidget(self._search_widget)

        name = 'Filters'
        icon = tp.ResourcesMgr().icon('filter', theme='black')
        icon.set_color(icon_color)
        tip = 'Filter the current results by type.\nCtrl + Click will hide the ohters and show the selected one.'
        self.add_menubar_action(name, icon, tip, callback=self._on_show_filter_by_menu)

        name = 'Item View'
        icon = tp.ResourcesMgr().icon('slider', theme='black')
        icon.set_color(icon_color)
        tip = 'Change the style of the item view'
        self.add_menubar_action(name, icon, tip, callback=self._on_show_item_view_menu)

        name = 'Group By'
        icon = tp.ResourcesMgr().icon('grid_view', theme='black')
        icon.set_color(icon_color)
        tip = 'Group the current items in the view by column'
        self.add_menubar_action(name, icon, tip, callback=self._on_show_group_by_menu)

        name = 'Sort By'
        icon = tp.ResourcesMgr().icon('descending_sorting', theme='black')
        icon.set_color(icon_color)
        tip = 'Sort the current items in the view by column'
        self.add_menubar_action(name, icon, tip, callback=self._on_show_sort_by_menu)

        name = 'View'
        icon = tp.ResourcesMgr().icon('add')
        # icon.set_color(icon_color)
        tip = 'Choose to show/hide both the preview and navigation pane\nCtrl + click will hide the menu bar as well.'
        self.add_menubar_action(name, icon, tip, callback=self._on_toggle_view)

        name = 'Sync Items'
        icon = tp.ResourcesMgr().icon('sync', theme='black')
        icon.set_color(icon_color)
        tip = 'Sync with the filesystem'
        self.add_menubar_action(name, icon, tip, callback=self._on_sync)

        name = 'Settings'
        icon = tp.ResourcesMgr().icon('settings', theme='black')
        icon.set_color(icon_color)
        tip = 'Settings menu'
        self.add_menubar_action(name, icon, tip, callback=self._on_show_settings_menu)

    # def event(self, ev):
    #     """
    #     Overrides window.MainWindow event function
    #     :param ev: QEvent
    #     :return: QEvent
    #     """
    #
    #     if isinstance(ev, QKeyEvent):
    #         if qtutils.is_control_modifier() and ev.key() == Qt.Key_F:
    #             self.search_widget().setFocus()
    #
    #     if isinstance(ev, QStatusTipEvent):
    #         self.status_widget().show_info_message(ev.tip())
    #
    #     return super(LibraryWindow, self).event(ev)

    def keyReleaseEvent(self, event):
        """
        Overrides window.MainWindow keyReleaseEvent function
        :param event: QKeyEvent
        """

        for item in self.selected_items():
            item.key_release_event(event)
        super(LibraryWindow, self).keyReleaseEvent(event)

    def show(self, **kwargs):
        """
        Overrides window.MainWindow show function
        We always raise_ the widget on show
        Use **kwargs to set platform dependent show options used in subclasses
        :param kwargs:
        :return:
        """

        # super(LibraryWindow, self).show(self)
        super(LibraryWindow, self).show()
        self.setWindowState(Qt.WindowNoState)
        self.raise_()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def library(self):
        """
        Returns muscle library
        :return: MuscleLibrary
        """

        return self._library

    def set_library(self, library):
        """
        Sets the muscle library
        :param library: MuscleLibrary
        """

        self._library = library

    def set_sizes(self, sizes):
        """
        Set sizes of window splitters
        :param sizes: list(int, int, int)
        """

        f_size, c_size, p_size = sizes
        p_size = p_size if p_size > 0 else 200
        f_size = f_size if f_size > 0 else 120
        self._splitter.setSizes([f_size, c_size, p_size])
        self._splitter.setStretchFactor(1, 1)

    def is_locked(self):
        """
        Returns whether the library is locked or not
        :return: bool
        """

        return self._is_locked

    def set_locked(self, flag):
        """
        Sets the state of the library to be editable or not
        :param flag: bool
        """

        self._is_locked = flag

        self.sidebar_widget().set_locked(flag)
        self.viewer().set_locked(flag)
        self.update_create_item_button()
        self.update_window_title()
        self.lockChanged.emit(flag)

    def superusers(self):
        """
        Returns the superusers for the window
        :return: list(str)
        """

        return self._superusers

    def set_superusers(self, superusers):
        """
        Sets the valid superusres for the library widget
        This will lock all folders unless the current user is a superuser
        :param superusers: list(str)
        """

        self._superusers = superusers
        self.update_lock()

    def lock_reg_exp(self):
        """
        Returns the lock regular expression used for locking the widget
        :return: str
        """

        return self._lock_reg_exp

    def set_lock_reg_exp(self, reg_exp):
        """
        Sets the lock regular expression used for locking the widget
        Locks only folders that contain the given regular expression in their path
        :param reg_exp: str
        """

        self._lock_reg_exp = reg_exp
        self.update_lock()

    def unlock_reg_exp(self):
        """
        Returns the unlock regular expression used for unlocking the widget
        :return: str
        """

        return self._unlock_reg_exp

    def set_unlock_reg_exp(self, reg_exp):
        """
        Sets the unlock regular expression used for locking the widget
        Unlocks only folders that contain the given regular expression in their path
        :param reg_exp: str
        """

        self._unlock_reg_exp = reg_exp
        self.update_lock()

    def is_lock_reg_exp_enabled(self):
        """
        Returns whether the lock regular expression or unlock regular expression has been set
        :return: bool
        """

        return not (self.superusers() is None and self.lock_reg_exp() is None and self.unlock_reg_exp() is None)

    def update_lock(self):
        """
        Updates the lock state for the library
        """

        if not self.is_lock_reg_exp_enabled():
            return

        superusers = self.superusers() or list()
        re_locked = re.compile(self.lock_reg_exp() or '')
        re_unlocked = re.compile(self.unlock_reg_exp() or '')

        print('Updating Lock ...')

    def is_refresh_enabled(self):
        """
        Returns whether refresh is enabled or not
        If not, all updates will be ignored
        :return: bool
        """

        return self._refresh_enabled

    def set_refresh_enabled(self, flag):
        """
        Whether widgets should be updated or not
        :param flag: bool
        """

        self.library().set_search_enabled(flag)
        self._refresh_enabled = flag

    def update(self):
        """
        Overrides base QMainWindow update function
        Update the library widget and the data
        """

        self.refresh_sidebar()
        self.update_window_title()

    def update_window_title(self):
        """
        Updates the window title
        """

        pass

    def name(self):
        """
        Return the name of the library
        :return: str
        """

        if not self._library:
            return

        return self._library.name()

    def path(self):
        """
        Returns the path being used by the library
        :return: str
        """

        if not self._library:
            return

        return self._library.path()

    def set_path(self, path):
        """
        Set the path being used by the library
        :param path: str
        """

        if path is None:
            path = ''
        else:
            path = path_utils.real_path(path)

        if path == self.path():
            qt.logger.warning('Path is already set!')
            return

        library = self.library()
        library.set_path(path)

        # if not os.path.exists(library.data_path()):
        #     library.sync()
        self.sync(force_start=True, force_searh=True)

        if self.stack.currentIndex() != 0:
            self.stack.slide_in_index(0)

        self.refresh()
        # self.library().search()
        self.update_preview_widget()

    def set_create_widget(self, create_widget):
        """
        Set the widget that should be showed when creating a new item
        :param create_widget: QWidget
        """

        if not create_widget:
            return

        self.set_new_item_widget_visible(True)
        # self.viewer().clear_selection()

        # fsize, rsize, psize = self._splitter.sizes()
        # if psize < 150:
        #     self.set_sizes((fsize, rsize, 180))

        # self.set_preview_widget(create_widget)
        self.set_new_item_widget(create_widget)
        self.stack.slide_in_index(1)

    def refresh(self):
        """
        Refresh all necessary items
        """

        if self.is_refresh_enabled():
            self.update()

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    # @show_wait_cursor_decorator
    # def load_settings(self):
    #     """
    #     Loads the user settings from disk
    #     """
    #
    #     self.reload_stylesheet()
    #     settings = self.read_settings()
    #     self.set_settings(settings)
    #
    # def read_settings(self):
    #     """
    #     Reads settings from dsik
    #     :return: dict
    #     """
    #
    #     key = self.name()
    #     return self.settings().get(key, {})
    #
    # def save_settings(self, settings=None):
    #     """
    #     Save the settings to the settings path
    #     :param settings: dict
    #     """
    #
    #     settings = settings or self.settings()
    #     key = self.name()
    #     self.show_toast_message('Saved')

    # def settings(self):
    #     """
    #     Return a dictionary with the widget settings
    #     :return: dict
    #     """
    #
    #     settings = dict()
    #
    #     settings['kwargs'] = self._kwargs
    #
    #     if self.theme():
    #         settings['theme'] = self.theme().settings()
    #
    #     settings['library'] = self.library().settings()
    #
    #     settings['viewerWidget'] = self.viewer().settings()
    #     settings['searchWidget'] = self.search_widget().settings()
    #     settings['sidebarWidget'] = self.sidebar_widget().settings()
    #
    #     settings['filterByMenu'] = self._filter_by_menu.settings()
    #
    #     settings['path'] = self.path()
    #
    #     return settings

    def save_settings(self, settings=None):

        settings = settings or self.settings()
        if not settings:
            return

        library_name = self.objectName().upper()
        self.settings().set('{}/path'.format(library_name), self.path())
        self.settings().set('{}/dpi'.format(library_name), self.dpi())
        self.settings().set('{}/panelSizes'.format(library_name), self._splitter.sizes())
        self.settings().set('{}/sidebarWidgetVisible'.format(library_name), self.is_sidebar_widget_visible())
        self.settings().set('{}/menuBarWidgetVisible'.format(library_name), self.is_menubar_widget_visible())
        self.settings().set('{}/previewWidgetVisible'.format(library_name), self.is_menubar_widget_visible())
        self.settings().set('{}/statusBarWidgetVisible'.format(library_name), self.is_menubar_widget_visible())
        self.settings().set('{}/searchWidget'.format(library_name), self.search_widget().settings())
        self.settings().set('{}/recursiveSearchEnabled'.format(library_name), self.is_recursive_search_enabled())
        self.settings().set('{}/filterByMenu'.format(library_name), self._filter_by_menu.settings())
        self.settings().set('{}/library'.format(library_name), self.library().settings())
        self.settings().set('{}/trashFolderVisible'.format(library_name), self.is_trash_folder_visible())
        self.settings().set('{}/viewerWidget'.format(library_name), self.viewer().settings())

    def set_settings(self, settings):
        """
        Set the library window settings from the given dictionary
        :param settings: dict
        """

        super(LibraryWindow, self).set_settings(settings)

        is_refresh_enabled = self.is_refresh_enabled()
        library_name = self.objectName().upper()

        try:
            self.set_refresh_enabled(False)
            self.viewer().set_toast_enabled(False)

            # if not self.path():
            #     path = self.settings().get('{}/path'.format(library_name))
            #     if path and os.path.exists(path):
            #         self.set_path(path)

            dpi = float(self.settings().get('{}/dpi'.format(library_name), 1.0))
            self.set_dpi(dpi)

            sizes = self.settings().get('{}/panelSizes'.format(library_name), [160, 280, 180])
            sizes = [int(i) for i in sizes]
            if sizes and len(sizes) == 3:
                self.set_sizes(sizes)

            sidebar_visible = bool(self.settings().get('{}/sidebarWidgetVisible'.format(library_name), True))
            if sidebar_visible:
                self.set_sidebar_widget_visible(sidebar_visible)

            menubar_visible = bool(self.settings().get('{}/menuBarWidgetVisible'.format(library_name), True))
            if menubar_visible:
                self.set_menubar_widget_visible(menubar_visible)

            preview_visible = bool(self.settings().get('{}/previewWidgetVisible'.format(library_name), True))
            if preview_visible:
                self.set_preview_widget_visible(preview_visible)

            status_visible = bool(self.settings().get('{}/statusBarWidgetVisible'.format(library_name), True))
            if status_visible:
                self.set_status_widget_visible(status_visible)

            search_widget = dict(self.settings().get('{}/searchWidget'.format(library_name), {'text': ''}))
            if search_widget:
                self.search_widget().set_settings(search_widget)

            recursive_search = bool(self.settings().get('{}/recursiveSearchEnabled'.format(library_name), True))
            if recursive_search:
                self.set_recursive_search_enabled(recursive_search)

            filter_by_menu = dict(self.settings().get('{}/filterByMenu'.format(library_name), {'Folder': False}))
            if filter_by_menu:
                self._filter_by_menu.set_settings(filter_by_menu)
        finally:
            self.reload_stylesheet()
            self.set_refresh_enabled(is_refresh_enabled)
            self.refresh()

        library_settings = dict(self.settings().get(
            '{}/library'.format(library_name), {'sortBy': ['name:asc'], 'groupBy': ['category:asc']}))
        if library_settings:
            self.library().set_settings(library_settings)

        trash_folder_visible = bool(self.settings().get(
            '{}/trashFolderVisible'.format(library_name), False))
        if trash_folder_visible:
            self.set_trash_folder_visible(trash_folder_visible)

        viewer_settings = dict(self.settings().get(
            '{}/viewerWidget'.format(library_name),
            {'spacing': 2, 'padding': 6, 'zoomAmount': 80, 'textVisible': True}))
        if viewer_settings:
            self.viewer().set_settings(viewer_settings)

        self.viewer().set_toast_enabled(True)

        self.update_filters_button()

    def update_settings(self, settings):
        """
        Save the given path to the settins on disk
        :param settings: dict
        """

        data = self.read_settings()
        data.update(settings)
        self.save_settings(data)

    def reset_settings(self):
        """
        Reset the settings to the default settings
        """

        self.set_settings(self.DEFAULT_SETTINGS)

    # ============================================================================================================
    # DRAG & DROP
    # ============================================================================================================

    def save_custom_order(self):
        """
        Function used for save the custom order
        """

        self.library().save_item_data(self.items(), emit_data_changed=True)

    def is_custom_order_enabled(self):
        """
        Returns whether custom order is enabled or not
        :return: bool
        """

        return 'Custom Order' in str(self.library().sort_by())

    def is_move_items_enabled(self):
        """
        Returns whether moving items via drag and drop is enabled or not
        :return: bool
        """

        paths = self.selected_folder_paths()
        if len(paths) != 1:
            return False
        if self.selected_items():
            return False

        return True

    def create_move_items_dialog(self):
        """
        Creates and returns a dialog for moving items
        :return: MessageBox
        """

        dlg = messagebox.create_message_box(
            self, 'Move or Copy Items?', 'Would you like to copy or move the selected item/s?')
        dlg.button_box().clear()
        dlg.add_button('Copy', QDialogButtonBox.AcceptRole)
        dlg.add_button('Move', QDialogButtonBox.AcceptRole)
        dlg.add_button('Cancel', QDialogButtonBox.RejectRole)

        return dlg

    def show_move_items_dialog(self, items, target_folder):
        """
        Shows the move items dialog for the given items
        :param items: list(LibraryItem)
        :param dst: target_folder
        """

        Copy = 0
        Cancel = 2

        for item in items:
            if item.dirname() == target_folder:
                return

        dlg = self.create_move_items_dialog()
        res = dlg.exec_()
        dlg.close()

        if res == Cancel:
            return

        copy = res == Copy

        self.move_items(items, target_folder, copy=copy)

    def move_items(self, items, target_folder, copy=False, force=False):
        """
        Moves the given items to the target folder path
        :param items: list(LibraryItem)
        :param target_folder: str
        :param copy: bool
        :param force: bool
        """

        self.viewer().clear_selection()
        moved_items = list()
        try:
            self.library().blockSignals(True)
            for item in items:
                path = target_folder + '/' + item.name()
                if force:
                    path = utils.generate_unique_path(path)
                if copy:
                    item.copy(path)
                else:
                    item.rename(path)
                moved_items.append(item)
        except Exception as e:
            self.show_exception_dialog('Move Error', e)
        finally:
            self.library().blockSignals(False)

            self.refresh()
            self.select_items(moved_items)
            self.scroll_to_selected_item()

    # ============================================================================================================
    # VIEWER WIDGET
    # ============================================================================================================

    def viewer(self):
        """
        Returns muscle viewer widget
        :return:
        """

        return self._viewer

    def add_item(self, item, select=False):
        """
        Add the given item to the viewer widget
        :param item: LibraryItem
        :param select: bool
        """

        self.add_items([item], select=select)

    def add_items(self, items, select=False):
        """
        Add the given items to the viewer widget
        :param items: list(LibraryItem)
        :param select: bool
        """

        self.viewer().add_items(items)
        self._items.extend(items)

        if select:
            self.select_Items(items)
            self.scroll_to_selected_item()

    def create_items_from_urls(self, urls):
        """
        Return a new list of items from the given urls
        :param urls: list(QUrl)
        :return: list(LibraryItem)
        """

        return self.manager().items_from_urls(urls, library_window=self)

    def items(self):
        """
        Return all the loaded items
        :return: list(LibraryItem)
        """

        return self._items

    def selected_items(self):
        """
        Return selected items
        :return: list(LibraryItem)
        """

        return self.viewer().selected_items()

    def select_path(self, path):
        """
        Select the item with the given path
        :param path: str
        """

        self.select_paths([path])

    def select_paths(self, paths):
        """
        Select the items with the given paths
        :param paths: list(str)
        """

        selection = self.selected_items()
        self.clear_preview_widget()
        self.viewer().clear_selection()
        self.viewer().select_paths(paths)

        if self.selected_items() != selection:
            self._on_item_selection_changed()

    def select_items(self, items):
        """
        Select the given items
        :param items: list(LibraryItem)
        :return:
        """

        paths = [item.path() for item in items]
        self.select_paths(paths)

    def scroll_to_selected_item(self):
        """
        Scroll the item widget to the selected item
        """

        self.viewer().scroll_to_selected_item()

    def refresh_selection(self):
        """
        Refresh teh current item selection
        """

        items = self.selected_items()
        self.viewer().clear_selection()
        self.select_items(items)

    def clear_items(self):
        """
        Remove all the loaded items
        """

        self.viewer().clear()

    # ============================================================================================================
    # MENUBAR WIDGET
    # ============================================================================================================

    def menubar_widget(self):
        """
        Returns menu bar widget
        :return: LibraryMenuBarWidget
        """

        return self._menubar_widget

    def add_menubar_action(self, name, icon, tip, callback=None):
        """
        Add a button/action into menu bar widget
        :param name: str
        :param icon: QIcon
        :param tip: str
        :param callback: fn
        :return: QAction
        """

        # We need to do this to avoid PySide2 errors
        def _callback():
            callback()

        action = self.menubar_widget().addAction(name)
        if icon:
            action.setIcon(icon)
        if tip:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if callback:
            action.triggered.connect(_callback)

        return action

    def is_menubar_widget_visible(self):
        """
        Returns whether MenuBar widget is visible or not
        :return: bool
        """

        return self.menubar_widget().is_expanded()

    def set_menubar_widget_visible(self, flag):
        """
        Sets whether menubar widget is visible or not
        :param flag: bool
        """

        flag = bool(flag)
        if flag:
            self.menubar_widget().expand()
        else:
            self.menubar_widget().collapse()

    # ============================================================================================================
    # SIDEBAR WIDGET
    # ============================================================================================================

    def sidebar_widget(self):
        """
        Return the sidebar widget
        :return: LibrarySidebarWidget
        """

        return self._sidebar_widget

    def is_sidebar_widget_visible(self):
        """
        Return whether SideBar widget is visible or not
        :return: bool
        """

        return self._sidebar_widget_visible

    def set_sidebar_widget_visible(self, flag):
        """
        Set whether SideBar widget is visible or not
        :param flag: bool
        """

        flag = bool(flag)
        self._sidebar_widget_visible = flag

        if flag:
            self._sidebar_frame.show()
        else:
            self._sidebar_frame.hide()

        self.update_view_button()

    @show_wait_cursor_decorator
    def refresh_sidebar(self):
        """
        Refresh the state of the sidebar widget
        """

        path = self.path()
        if not path and self._allow_non_path:
            return
        else:
            if not path:
                return self.show_hello_dialog()
            elif not os.path.exists(path):
                return self.show_path_error_dialog()

            self.update_sidebar()

    def update_sidebar(self):
        """
        Update the folders to be shown in the folders widget
        """

        data = dict()
        root = self.path()

        queries = [{'filters': [('type', 'is', 'Folder')]}]

        items = self.library().find_items(queries)
        trash_icon_path = tp.ResourcesMgr().get('icons', 'black', 'trash')

        for item in items:
            path = item.path()
            if item.path().endswith('Trash'):
                data[path] = {'iconPath': trash_icon_path}
            else:
                data[path] = dict()

        self.sidebar_widget().set_data(data, root=root)

    def selected_folder_path(self):
        """
        Return the selected folder items
        :return: str or None
        """

        return self.sidebar_widget().selected_path()

    def selected_folder_paths(self):
        """
        Return the selected folder items
        :return: list(str)
        """

        return self.sidebar_widget().selected_paths()

    def select_folder_path(self, path):
        """
        Select the given folder path
        :param path: str
        """

        self.select_folder_paths([path])

    def select_folder_paths(self, paths):
        """
        Select the given folder paths
        :param paths: list(str)
        """

        self.sidebar_widget().select_paths(paths)

    def create_folder_context_menu(self):
        """
        Returns the folder menu for the selected folders
        :return: QMenu
        """

        path = self.selected_folder_path()
        items = list()
        if path:
            item = self.manager().item_from_path(path, library_window=self)
            if item:
                items = [item]

        return self._create_item_context_menu(items)

    # ============================================================================================================
    # NEW ITEM WIDGET
    # ============================================================================================================

    def new_item_widget(self):
        """
        Returns the current new item widget
        :return: QWidget
        """

        return self._new_item_widget

    def set_new_item_widget(self, widget):
        """
        Sets the new item widget
        :param widget: QWidget
        """

        if self._new_item_widget == widget:
            msg = 'New Item widget already contains widghet {}'.format(widget)
            qt.logger.debug(msg)
        else:
            self.close_new_item_widget()
            self._new_item_widget = widget
            if self._new_item_widget:
                self._new_item_frame.layout().addWidget(self._new_item_widget)
                self._new_item_widget.show()

    def is_new_item_widget_visible(self):
        """
        Returns whether new item widget is visible or not
        :return: bool
        """

        return self._new_item_widget_visible

    def set_new_item_widget_visible(self, flag):
        """
        Set if the New Item widget should be showed or not
        :param flag: bool
        """

        flag = bool(flag)
        self._new_item_widget_visible = flag

        if flag:
            self.stack.slide_in_index(1)
            self._new_item_frame.show()
        else:
            self.stack.slide_in_index(0)
            self._new_item_frame.hide()

        self.update_view_button()

    def clear_new_item_widget(self):
        """
        Set the default new item widget
        """

        self._new_item_widget = None
        widget = base.PlaceholderWidget()
        self.set_new_item_widget(widget)

    def close_new_item_widget(self):
        """
        Close and delete the new item widget
        """

        lyt = self._new_item_frame.layout()

        while lyt.count():
            item = lyt.takeAt(0)
            item.widget().hide()
            item.widget().close()
            item.widget().deleteLater()

        self._new_item_widget = None

    # ============================================================================================================
    # PREVIEW WIDGET
    # ============================================================================================================

    def preview_widget(self):
        """
        Returns the current preview widget
        :return: QWidget
        """

        return self._preview_widget

    def set_preview_widget(self, widget):
        """
        Set the preview widget
        :param widget: QWidget
        """

        if self._preview_widget == widget:
            msg = 'Preview widget already contains widget {}'.format(widget)
            qt.logger.debug(msg)
        else:
            self.close_preview_widget()
            self._preview_widget = widget
            if self._preview_widget:
                self._preview_frame.layout().addWidget(self._preview_widget)
                self._preview_widget.show()

    def is_preview_widget_visible(self):
        """
        Returns whether preview widget is visible or not
        :return: bool
        """

        return self._preview_widget_visible

    def set_preview_widget_visible(self, flag):
        """
        Set if the PreviewWidget should be showed or not
        :param flag: bool
        """

        flag = bool(flag)
        self._preview_widget_visible = flag

        if flag:
            self._preview_frame.show()
        else:
            self._preview_frame.hide()

        self.update_view_button()

    def update_preview_widget(self):
        """
        Update the current preview widget
        """

        self.set_preview_widget_from_item(self._current_item, force=True)

    def set_preview_widget_from_item(self, item, force=True):
        """
        Set the preview widget from the given item
        :param item: LibvraryItem
        :param force: bool
        """

        if not force and self._current_item == item:
            qt.logger.debug('The current item preview widget is already set!')
            return

        self._current_item = item

        if item:
            self.close_preview_widget()
            try:
                item.show_preview_widget(self)
            except Exception as e:
                self.show_error_message(e)
                self.clear_preview_widget()
                raise
        else:
            self.clear_preview_widget()

    def clear_preview_widget(self):
        """
        Set the default preview widget
        """

        self._preview_widget = None
        widget = base.PlaceholderWidget()
        self.set_preview_widget(widget)

    def close_preview_widget(self):
        """
        Close and delete the preview widget
        """

        lyt = self._preview_frame.layout()

        while lyt.count():
            item = lyt.takeAt(0)
            item.widget().hide()
            item.widget().close()
            item.widget().deleteLater()

        self._preview_widget = None

    # ============================================================================================================
    # SEARCH WIDGET
    # ============================================================================================================

    def search_widget(self):
        """
        Returns search widget
        :return: LibrarySearchWidget
        """

        return self._search_widget

    def set_search_text(self, text):
        """
        Set the search widget text
        :param text: str
        """

        self.search_widget().setText(text)

    def items_visible_count(self):
        """
        Return the number of visible items
        :return: int
        """

        return self._items_visible_count

    def items_hidden_count(self):
        """
        Return the number of hidden items
        :return: int
        """

        return self._items_hidden_count

    # ============================================================================================================
    # STATUS WIDGET
    # ============================================================================================================

    def status_widget(self):
        """
        Returns the status widget
        :return: StatusWidget
        """

        return self._status_widget

    def is_status_widget_visible(self):
        """
        Return whether StatusWidget is visible or not
        :return: bool
        """

        return self._status_widget_visible

    def set_status_widget_visible(self, flag):
        """
        Set whether StatusWidget is visible or not
        :param flag: bool
        """

        flag = bool(flag)
        self._status_widget_visible = flag
        if flag:
            self.status_widget().show()
        else:
            self.status_widget().hide()

    # ============================================================================================================
    # TRASH FOLDER
    # ============================================================================================================

    def trash_enabled(self):
        """
        Returns True if moving items to trash
        :return: bool
        """

        return self._trash_enabled

    def set_trash_enabled(self, flag):
        """
        Sets if items can be trashed or not
        :param flag: bool
        """

        self._trash_enabled = flag

    def is_path_in_trash(self, path):
        """
        Returns whether given path is in trash or not
        :param path: str
        :return: bool
        """

        return consts.TRASH_NAME in path.lower()

    def trash_path(self):
        """
        Returns the trash path for the library
        :return: str
        """

        path = self.path()
        return '{}/{}'.format(path, consts.TRASH_NAME.title())

    def trash_folder_exists(self):
        """
        Returns whether trash folder exists or not
        :return: bool
        """

        return os.path.exists(self.trash_path())

    def create_trash_folder(self):
        """
        Create the trash folder if it does not already exists
        """

        trash_path = self.trash_path()
        if not os.path.exists(trash_path):
            os.makedirs(trash_path)

    def is_trash_folder_visible(self):
        """
        Return whether trash folder is visible or not
        :return: bool
        """

        return self._is_trash_folder_visible

    def set_trash_folder_visible(self, flag):
        """
        Set whether trash folder is visible or not
        :param flag: bool
        """

        self._is_trash_folder_visible = flag

        if flag:
            query = {
                'name': 'trash_query',
                'filters': list()
            }
        else:
            query = {
                'name': 'trash_query',
                'filters': [('path', 'not_contains', 'Trash')]
            }

        self.library().add_to_global_query(query)
        self.update_sidebar()
        self.library().search()

    def is_trash_selected(self):
        """
        Return whether the selected folders are in the trash or not
        :return: bool
        """

        folders = self.selected_folder_paths()
        for folder in folders:
            if self.is_path_in_trash(folder):
                return True

        items = self.selected_items()
        for item in items:
            if self.is_path_in_trash(item.path()):
                return True

        return False

    def move_items_to_trash(self, items):
        """
        Move the given items to trash path
        :param items: list(LibraryItem)
        """

        self.create_trash_folder()
        self.move_items(items, target_folder=self.trash_path(), force=True)

    def show_move_items_to_trash_dialog(self, items=None):
        """
        Show the "Move to Trash" dialog for the selected items
        :param items: list(LibraryItem) or None
        """

        items = items or self.selected_items()
        if items:
            title = 'Move to Trash?'
            text = 'Are you sure you want to move the selected item/s to the trash?'
            result = self.show_question_dialog(title, text)
            if result == QDialogButtonBox.Yes:
                self.move_items_to_trash(items)

    # ============================================================================================================
    # DPI SUPPORT
    # ============================================================================================================

    def dpi(self):
        """
        Return the current dpi for the library widget
        :return: float
        """

        return float(self._dpi)

    def set_dpi(self, dpi):
        """
        Set the current dpi for the library widget
        :param dpi: float
        """

        if not consts.DPI_ENABLED:
            dpi = 1.0

        self._dpi = dpi

        self.viewer().set_dpi(dpi)
        self.menubar_widget().set_dpi(dpi)
        self.sidebar_widget().set_dpi(dpi)
        self.status_widget().setFixedHeight(20 * dpi)

        self._splitter.setHandleWidth(2 * dpi)

        self.show_toast_message('DPI: {}'.format(int(dpi * 100)))

        self.reload_stylesheet()

    @show_wait_cursor_decorator
    def sync(self, force_start=False, force_searh=False):
        """
        Sync any data that might be out of date with the model
        """

        @show_wait_cursor_decorator
        def _sync():
            elapsed_time = time.time()
            self.library().sync(percent_callback=self.set_progress_bar_value)
            elapsed_time = time.time() - elapsed_time
            self.status_widget().show_info_message('Synced items in {0:.3f} seconds'.format(elapsed_time))
            self.set_progress_bar_value('Done')
            if force_start:
                progress_bar.close()
            else:
                animation.fade_out_widget(progress_bar, duration=500, on_finished=progress_bar.close)

            if force_searh and self.library():
                self.library().set_dirty(True)
                self.library().search()

        progress_bar = self.status_widget().progress_bar()
        if self.PROGRESS_BAR_VISIBLE:
            progress_bar.show()
        self.set_progress_bar_value('Syncing')

        if force_start:
            _sync()
        else:
            animation.fade_in_widget(progress_bar, duration=1, on_finished=_sync)

    def set_progress_bar_value(self, label, value=-1):
        """
        Sets the progress bar label and value
        :param label: str
        :param value: int
        """

        progress_bar = self.status_widget().progress_bar()
        if value == -1:
            self.status_widget().progress_bar().reset()
            progress_bar.set_value(100)
            progress_bar.set_text(label)
        else:
            percent = value * 100.0
            text = '{0} {1:.0f}%'.format(label, percent)
            progress_bar.set_value(percent)
            progress_bar.set_text(text)

    # ============================================================================================================
    # THEMES/STYLES
    # ============================================================================================================

    def icon_color(self):
        """
        Returns the icon color
        :return: Color
        """

        return consts.ICON_COLOR

    def reload_stylesheet(self):
        """
        Reloads the style to the current theme
        """

        pass

    # ============================================================================================================
    # TOAST/MESSAGES
    # ============================================================================================================

    def show_toast_message(self, text, duration=1000):
        """
        Shows toast widget with the given text and duration
        :param text: str
        :param duration:int
        """

        self.viewer().show_toast_message(text, duration)

    def show_info_message(self, text, msecs=None):
        """
        Shows info message to the user
        :param text: str
        :param msecs: int or None
        """

        self.status_widget().show_info_message(text)

    def show_warning_message(self, text, msecs=None):
        """
        Shows warning message to the user
        :param text: str
        :param msecs: int or None
        """

        self.status_widget().show_warning_message(text, msecs)
        self.set_status_widget_visible(True)

    def show_error_message(self, text, msecs=None):
        """
        Shows error message to the user
        :param text:str
        :param msecs: int or None
        """

        self.status_widget().show_error_message(text, msecs)
        self.set_status_widget_visible(True)

    def show_refresh_message(self):
        """
        Show long the current refresh took
        """

        item_count = len(self.library().results())
        elapsed_time = self.library().search_time()

        plural = ''
        if item_count > 1:
            plural = 's'

        msg = 'Found {0} item{1} in {2:.3f} seconds.'.format(item_count, plural, elapsed_time)
        self.status_widget().show_info_message(msg)

        qt.logger.debug(msg)

    @show_arrow_cursor_decorator
    def show_hello_dialog(self):
        """
        This function is called when there is not root path set for the library
        """

        text = 'Please choose a folder location for storing the data'
        dialog = messagebox.create_message_box(
            None, 'Welcome {}'.format(self.name()), text, header_pixmap=self._library_icon_path)
        dialog.accepted.connect(self.show_change_path_dialog)
        dialog.exec_()

    @show_arrow_cursor_decorator
    def show_path_error_dialog(self):
        """
        This function is called when the root path does not exists during refresh
        """

        path = self.path()
        text = 'The current root path does not exists "{}". Please select a new root path to continue.'.format(path)
        dialog = messagebox.create_message_box(self, 'Path Error', text)
        dialog.show()
        dialog.accepted.connect(self.show_change_path_dialog)

    def show_change_path_dialog(self):
        """
        Shows a file browser dialog for changing the root path
        :return: str
        """

        path = self._show_change_path_dialog()
        if path:
            self.set_path(path)
        else:
            self.refresh()

    def show_info_dialog(self, title, text):
        """
        Function that shows an information dialog to the user
        :param title: str
        :param text: str
        :return: QDialogButtonBox.StandardButton
        """

        buttons = QDialogButtonBox.Ok
        return messagebox.MessageBox.question(self, title, text, buttons=buttons)

    def show_question_dialog(self, title, text, buttons=None):
        """
        Function that shows a question dialog to the user
        :param title: str
        :param text: str
        :param buttons: list(QDialogButtonBox.StandardButton)
        :return: QDialogButtonBox.StandardButton
        """

        buttons = buttons or QDialogButtonBox.Yes | QDialogButtonBox.No | QDialogButtonBox.Cancel
        return messagebox.MessageBox.question(self, title, text, buttons=buttons)

    def show_error_dialog(self, title, text):
        """
        Function that shows an error dialog to the user
        :param title: str
        :param text: str
        :return: QDialogButtonBox.StandardButton
        """

        self.show_error_message(text)
        return messagebox.MessageBox.critical(self, title, text)

    def show_exception_dialog(self, title, text):
        """
        Function that shows an exception dialog to the user
        :param title: str
        :param text: str
        :return: QDialogButtonBox.StandardButton
        """

        qt.logger.exception(text)
        self.show_error_dialog(title, text)

    def show_folder_menu(self, pos=None):
        """
        Shows the folder context menu at the current cursor position
        :param pos: variant, QPoint or None
        :return: QAction
        """

        menu = self.create_folder_context_menu()
        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)
        action = menu.exec_(point)
        menu.close()

        return action

    # ============================================================================================================
    # OTHERS
    # ============================================================================================================

    def is_compact_view(self):
        """
        Returns whether the folder and preview widget are hidden
        :return: bool
        """

        return not self.is_sidebar_widget_visible() and not self.is_preview_widget_visible()

    def toggle_view(self):
        """
        Toggles the preview widget and folder widget visibility
        """

        compact = self.is_compact_view()
        if qtutils.is_control_modifier():
            compact = False
            self.set_menubar_widget_visible(compact)

        self.set_preview_widget_visible(compact)
        self.set_sidebar_widget_visible(compact)

    def update_view_button(self):
        """
        Updates the icon for the view action
        """

        compact = self.is_compact_view()
        action = self.menubar_widget().find_action('View')
        if not compact:
            icon = tp.ResourcesMgr().icon('view_all', theme='black')
        else:
            icon = tp.ResourcesMgr().icon('view_compact', theme='black')
        icon.set_color(self.icon_color())
        action.setIcon(icon)

    def update_filters_button(self):
        """
        Updates the icon for the filters menu
        """

        action = self.menubar_widget().find_action('Filters')
        icon = tp.ResourcesMgr().icon('filter', theme='black')
        icon.set_color(self.icon_color())
        if self._filter_by_menu.is_active():
            icon.set_badge(18, 1, 9, 9, color=consts.ICON_BADGE_COLOR)
        action.setIcon(icon)

    def is_recursive_search_enabled(self):
        """
        Returns whether recursive search is enabled or not
        :return: bool
        """

        return self.sidebar_widget().is_recursive()

    def set_recursive_search_enabled(self, flag):
        """
        Sets whether recursive search is enabled or not
        :param flag: bool
        """

        self.sidebar_widget().set_recursive(flag)

    # ============================================================================================================
    # INTERNAL FUNCTIONS
    # ============================================================================================================

    def _get_add_icon(self, color):
        """
        Returns add icon
        :param color: QColor
        :return: QIcon
        """

        return tp.ResourcesMgr().icon('add')

    def _create_new_item_menu(self):
        """
        Internal function that creates a new item menu for adding new items
        :return: QMenu
        """

        color = self.icon_color()

        item_icon = self._get_add_icon(color=color)
        menu = QMenu(self)
        menu.setIcon(item_icon)
        menu.setTitle('New')

        for cls in sorted(self.manager().registered_items(), key=operator.attrgetter('MenuOrder')):
            action = cls.create_action(menu, self)
            if action:
                action_icon = icon.Icon(action.icon())
                # action_icon.set_color(self.icon_color())
                action.setIcon(action_icon)
                menu.addAction(action)

        return menu

    def _create_item_context_menu(self, items):
        """
        Internal function that returns the item context menu for the given items
        :param items: list(LibraryItem)
        :return:
        """

        context_menu = menu.Menu(self)

        item = None
        if items:
            item = items[-1]
            item.context_menu(context_menu)

        if not self.is_locked():
            context_menu.addMenu(self._create_new_item_menu())
            if item:
                edit_icon = tp.ResourcesMgr().icon('edit')
                edit_menu = menu.Menu(context_menu)
                edit_menu.setTitle('Edit')
                edit_menu.setIcon(edit_icon)
                context_menu.addMenu(edit_menu)
                item.context_edit_menu(edit_menu)
                if self.trash_enabled():
                    edit_menu.addSeparator()
                    callback = partial(self.show_move_items_to_trash_dialog, items)
                    action = QAction('Move to Trash', edit_menu)
                    action.setEnabled(not self.is_trash_selected())
                    action.triggered.connect(callback)
                    edit_menu.addAction(action)

        context_menu.addSeparator()
        context_menu.addMenu(self._create_settings_menu())

        return context_menu

    def _create_settings_menu(self):
        """
        Returns the settings menu for changing the lirary widget
        :return: QMenu
        """

        settings_icon = tp.ResourcesMgr().icon('settings')
        context_menu = menu.Menu('', self)
        context_menu.setTitle('Settings')
        context_menu.setIcon(settings_icon)

        if consts.SETTINGS_DIALOG_ENABLED:
            settings_action = context_menu.addAction('Settings')
            settings_action.triggered.connect(self._on_show_settings_dialog)

        sync_action = context_menu.addAction('Sync')
        sync_action.triggered.connect(self.sync)

        context_menu.addSeparator()

        if consts.DPI_ENABLED:
            dpi_action = action.SliderAction('Dpi', context_menu)
            dpi = self.dpi() * 100.0
            dpi_action.slider().setRange(consts.DPI_MIN_VALUE, consts.DPI_MAX_VALUE)
            dpi_action.slider().setValue(dpi)
            dpi_action.valueChanged.connect(self._on_dpi_slider_changed)
            context_menu.addAction(dpi_action)
            context_menu.addSeparator()

        show_menu_action = QAction('Show Menu', context_menu)
        show_menu_action.setCheckable(True)
        show_menu_action.setChecked(self.is_menubar_widget_visible())
        show_menu_action.triggered[bool].connect(self.set_menubar_widget_visible)
        context_menu.addAction(show_menu_action)

        show_sidebar_action = QAction('Show Sidebar', context_menu)
        show_sidebar_action.setCheckable(True)
        show_sidebar_action.setChecked(self.is_sidebar_widget_visible())
        show_sidebar_action.triggered[bool].connect(self.set_sidebar_widget_visible)
        context_menu.addAction(show_sidebar_action)

        show_preview_action = QAction('Show Preview', context_menu)
        show_preview_action.setCheckable(True)
        show_preview_action.setChecked(self.is_preview_widget_visible())
        show_preview_action.triggered[bool].connect(self.set_preview_widget_visible)
        context_menu.addAction(show_preview_action)

        show_status_action = QAction('Show Status', context_menu)
        show_status_action.setCheckable(True)
        show_status_action.setChecked(self.is_status_widget_visible())
        show_status_action.triggered[bool].connect(self.set_status_widget_visible)
        context_menu.addAction(show_status_action)

        context_menu.addSeparator()

        if self.trash_enabled():
            show_trash_action = QAction('Show Trash Folder', context_menu)
            show_trash_action.setEnabled(self.trash_folder_exists())
            show_trash_action.setCheckable(True)
            show_trash_action.setChecked(self.is_trash_folder_visible())
            show_trash_action.triggered[bool].connect(self.set_trash_folder_visible)
            context_menu.addAction(show_trash_action)

        context_menu.addSeparator()

        recursive_search_action = QAction('Enable Recrusive Search', context_menu)
        recursive_search_action.setCheckable(True)
        recursive_search_action.setChecked(self.is_recursive_search_enabled())
        recursive_search_action.triggered[bool].connect(self.set_recursive_search_enabled)
        context_menu.addAction(recursive_search_action)

        context_menu.addSeparator()

        view_menu = self.viewer().create_settings_menu()
        context_menu.addMenu(view_menu)

        # context_menu.addSeparator()
        #
        # debug_mode_action = QAction('Debug Mode')
        # debug_mode_action.setCheckable(True)
        # debug_mode_action.setChecked(self.is_debug())
        # debug_mode_action.triggered[bool].connect(self.set_debug_mode)
        # context_menu.addAction(debug_mode_action)

        return context_menu

    @show_arrow_cursor_decorator
    def _show_change_path_dialog(self):
        """
        Internal function that opens a file dialog for setting a new root path
        :return: str
        """

        path = self.path()
        directory = path
        if not directory:
            directory = os.path.expanduser('~')
        dialog = QFileDialog(None, Qt.WindowStaysOnTopHint)
        dialog.setWindowTitle('Choose the root location')
        dialog.setDirectory(directory)
        dialog.setFileMode(QFileDialog.DirectoryOnly)
        if dialog.exec_() == QFileDialog.Accepted:
            selected_files = dialog.selectedFiles()
            if selected_files:
                path = selected_files[0]

        if not path:
            return

        path = path_utils.normalize_path(path)

        return path

    # ============================================================================================================
    # CALLBACK FUNCTIONS
    # ============================================================================================================

    def _on_search_finished(self):
        self.show_refresh_message()

    def _on_show_items_context_menu(self, pos=None):
        """
        Internal callback function that is called when user right clicks on muscle viewer
        Shows the item context menu at the current cursor position
        :param pos, QPoint
        :return: QAction
        """

        items = self.viewer().selected_items()
        menu = self._create_item_context_menu(items)
        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)
        action = menu.exec_(point)

        return action

    def _on_show_settings_dialog(self):
        pass

    def _on_sync(self):
        """
        Internal callback function that is executed when the user selects the Sync
        context menu action
        """

        self.sync(force_searh=True)

    def _on_show_new_menu(self):
        """
        Internal callback function triggered when the user presses New Item Action
        Creates and shows the new menu at the new action button
        :return: QAction
        """

        menu = self._create_new_item_menu()
        point = self.menubar_widget().rect().bottomLeft()
        point = self.menubar_widget().mapToGlobal(point)
        menu.show()

        return menu.exec_(point)

    def _on_show_settings_menu(self):
        """
        Internal callback function triggered when the user show settings
        Show the settings menu at the current cursor position
        :return: QAction
        """

        menu = self._create_settings_menu()
        point = self.menubar_widget().rect().bottomRight()
        point = self.menubar_widget().mapToGlobal(point)
        menu.show()
        point.setX(point.x() - menu.width())

        return menu.exec_(point)

    def _on_show_filter_by_menu(self):
        """
        Internal callback function called when the user clicks on the filter action
        """

        widget = self.menubar_widget().find_tool_button('Filters')
        point = widget.mapToGlobal(QPoint(0, widget.height()))
        self._filter_by_menu.show(point)
        self.update_filters_button()

    def _on_show_item_view_menu(self):
        """
        Internal callback function when triggered when the user clicks on item view action
        """

        menu = self.viewer().create_item_settings_menu()
        item_view_widget = self.menubar_widget().find_tool_button('Item View')
        point = item_view_widget.mapToGlobal(QPoint(0, item_view_widget.height()))
        menu.exec_(point)

    def _on_show_group_by_menu(self):
        """
        Internal callback function called when the user presses group by action
        """

        widget = self.menubar_widget().find_tool_button('Group By')
        point = widget.mapToGlobal(QPoint(0, widget.height()))
        self._group_by_menu.show(point)

    def _on_show_sort_by_menu(self):
        """
        Internal callback function called when the user presses Sort by action
        """

        widget = self.menubar_widget().find_tool_button('Sort By')
        point = widget.mapToGlobal(QPoint(0, widget.height()))
        self._sort_by_menu.show(point)

    def _on_toggle_view(self):
        """
        Internal callback function called whe the user press view action
        """

        self.toggle_view()

    def _on_item_moved(self, item):
        """
        Internal callback function that is triggered when the custom order has changed
        :param item, LibraryItem
        """

        self.save_custom_order()

    def _on_item_dropped(self, event):
        """
        Internal callback function that is triggered when items are dropped on the viewer or sidebar widget
        :param event: QEvent
        """

        mime_data = event.mimeData()
        print(mime_data)

    def _on_item_selection_changed(self):
        """
        Internal callback function that is triggered when an item is selected or deselected
        """

        item = self.viewer().selected_item()

        self.set_preview_widget_from_item(item)
        self.itemSelectionChanged.emit(item)

    def _on_folder_selection_changed(self):
        """
        Internal callback function that is triggered when a folder is selected or deselected
        """

        path = self.selected_folder_path()
        self.library().search()
        self.folderSelectionChanged.emit(path)
        self.globalSignal.folderSelectionChanged.emit(self, path)

    def _on_show_folder_menu(self, pos=None):
        """
        Internal callback function that is triggered when the user left click on sidebar widget
        """

        self.show_folder_menu(pos=pos)

    def _on_dpi_slider_changed(self, value):
        """
        Internal callback function triggered when the dpi action slider changes its values
        :param value: float
        """

        dpi = value / 100.0
        self.set_dpi(dpi)
